#!/usr/bin/env python3

"""
This program takes an input file in the following format:
[n]
[x1] [y1]
[x2] [y2]
...
[xn] [yn]

The first line indicates the number of cities. Each city is a
point in the plane, and each subsequent line indicates the
x- and y-coordinates of a single city using floating-point values.

The distance between two cities is defined as the Euclidean distance.
"""

from argparse import ArgumentParser
from itertools import combinations
from math import sqrt, inf
from typing import List, Tuple

# Define a type for editor help (requires Python 3.9+)
City = Tuple[float, float]


def bellman_held_karp(cities: List[City]) -> float:
    """Compute the minimum tour using Bellman-Held-Karp.

    The Bellman-Held-Karp algorithm is a dynamic programming solution
    to the traveling salesperson problem with a running time of
    O(n^2 2^n).

    Args:
        cities (list[City]): List of all city coordinates

    Returns:
        float: cost of minimum tour
    """

    n = len(cities)

    # Compute a symmetric Euclidean distances matrix
    dists = [[0.0] * n for _ in range(n)]
    for i in range(n):
        x1, y1 = cities[i]
        for j in range(i + 1, n):
            x2, y2 = cities[j]
            dists[i][j] = dists[j][i] = sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    # Compute all one-hop distances
    costs = {}
    for c in range(1, n):
        costs[(frozenset({0, c}), c)] = dists[0][c]

    cities_without_first = range(1, n)

    # Compute minimum tour
    for size in range(2, n):

        for subset in combinations(cities_without_first, size):

            subset_without_first = set(subset)
            subset_with_first = set(subset) | {0}

            for nxt in subset_without_first:

                min_path_cost = inf
                for end in subset_without_first:

                    if end != nxt:
                        new_path_cost = (
                            costs[(frozenset(subset_with_first - {nxt}), end)]
                            + dists[nxt][end]
                        )

                        if new_path_cost < min_path_cost:
                            min_path_cost = new_path_cost

                costs[(frozenset(subset_with_first), nxt)] = min_path_cost

    # Add the original node to the distance for all complete tours
    tour_costs = [costs[(a, k)] + dists[0][k] for a, k in costs if len(a) == n]

    min_tour_cost = min(tour_costs)
    return min_tour_cost


def main():

    aparser = ArgumentParser("Run the Bellman-Held-Karp algorithm for TSP.")
    aparser.add_argument("file", help="A file with a list of city locations.")
    args = aparser.parse_args()

    # Read the input file data
    with open(args.file, "r") as tspfile:
        n = int(tspfile.readline())
        cities = [(float(l.split()[0]), float(l.split()[1])) for l in tspfile]

    print(f"Number of cities: {n}, len(cities) = {len(cities)}")
    print(f"Length of optimal tour: {bellman_held_karp(cities):.3f}")


if __name__ == "__main__":
    main()
